#!/usr/bin/env python3
import gurobipy as g
import matplotlib.pyplot as plt
import numpy as np

# data --------------------------------------------------
d = [5, 5, 5, 5, 5, 10, 10, 15, 20, 20, 30, 30, 40, 50, 60, 60, 60, 50, 40, 30, 30, 20, 10, 5]

T = len(d)
n_base, e_base, c_base = 10, 7, (2.0 / 24.0)
n_peak, e_peak, c_peak = 40, 2, 12
s_max = 100
gamma = 0.75

# model ------------------------------------------------
m = g.Model()

# - add variables
# The number of running base load power plants.
base = m.addVar(0, n_base, 0, g.GRB.INTEGER)

# How many peaking power plants are running in each hour.
peak = m.addVars(T, lb=0, ub=n_peak, vtype=g.GRB.INTEGER)

# Energy storage state in each hour.
s = m.addVars(T+1, lb=0, ub=s_max, vtype=g.GRB.CONTINUOUS)

# Amount of energy taken from the storage.
a_take = m.addVars(T, lb=0, ub=g.GRB.INFINITY, vtype=g.GRB.CONTINUOUS)

# Amount of energy put into the storage.
a_put = m.addVars(T, lb=0, ub=g.GRB.INFINITY, vtype=g.GRB.CONTINUOUS)

# Whether we are putting or taking from the storage.
y = m.addVars(T, vtype=g.GRB.BINARY)

# Amount of generated energy.
gt = m.addVars(T, lb=0, ub=g.GRB.INFINITY, vtype=g.GRB.CONTINUOUS)

# - add constraints
big_m = s_max
m.addConstr(s[0] == 0)
for t in range(T):
    m.addConstr(e_base * base + e_peak * peak[t] == gt[t])
    m.addConstr(gt[t] + gamma * a_take[t] - a_put[t] == d[t])
    m.addConstr(s[t + 1] == (s[t] - a_take[t]) + a_put[t])
    m.addConstr(big_m * (1 - y[t]) >= a_take[t])
    m.addConstr(big_m * y[t] >= a_put[t])

# - set objective
m.setObjective(24 * c_base * base + g.quicksum([c_peak * peak[t] for t in range(T)]), g.GRB.MINIMIZE)

# call the solver ----------------------------------
m.optimize()

# print the solution -------------------------------
print("x_base = ", base.X)
for t in range(T):
    print(t, ":", "x_peak(t) = ", peak[t].X)

# visualization ------------------------------------
# Demand plot.
y_lim_max = max(s_max, max([gt[t].x + a_take[t].x for t in range(T)])) * 1.1

margin = 0.2
width = 0.3
plt.figure(figsize=(10, 4))
plt.bar([t + margin for t in range(T)], d, width=width, color='yellow')

bottom = np.zeros(T)
g_base = np.array([e_base * base.x for t in range(T)])
plt.bar([t + margin + width for t in range(T)],
        g_base,
        width=width,
        bottom=bottom,
        color='red')
bottom += g_base

g_peak = np.array([e_peak * peak[t].x for t in range(T)])
plt.bar([t + margin + width for t in range(T)],
        g_peak,
        width=width,
        bottom=bottom,
        color='green')
bottom += g_peak

g_a_take = np.array([gamma * a_take[t].x for t in range(T)])
plt.bar([t + margin + width for t in range(T)],
        g_a_take,
        width=width,
        bottom=bottom,
        color='blue')
bottom += g_a_take

plt.xlabel("hour")
plt.ylabel("energy")
plt.legend(['demand', 'base', 'peak', 'storage'], ncol=1, loc=2)
plt.xlim(0, 24)
plt.ylim(0, y_lim_max)
plt.xticks(range(24), [i % 24 for i in range(24)])
plt.grid()
plt.savefig('../demand.eps', bbox_inches='tight')

# Storage plot.
plt.figure(figsize=(10, 4))
plt.xlim(0, 24)
plt.xticks(range(24), [i % 24 for i in range(24)])
plt.xlabel("hour")
plt.ylabel("stored energy")
plt.ylim(0, y_lim_max)
plt.bar([t + margin for t in range(T)], [s[t].x for t in range(T)], width=width, color='purple')
plt.grid()
plt.savefig('../storage.eps', bbox_inches='tight')

plt.show()